/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/


#include <fstream>
#include <set>

#include <cstdlib>

#include <crown/allocation.h>
#include <crown/mapping.h>
#include <crown/frequency.h>
#include <crown/crown.h>

#include <pelib/AmplInput.hpp>
#include <pelib/AmplOutput.hpp>
#include <pelib/Vector.hpp>
#include <pelib/Matrix.hpp>
#include <pelib/Set.hpp>

#include <crown/allocation.h>

#ifdef debug
#undef debug
#endif
#if 0
#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << var << "\"" << endl;
#else
#define debug(var)
#endif

using namespace pelib;
using namespace std;

namespace pelib
{
namespace crown
{

Algebra
neighbor_allocation(const Algebra &solution, float tasks, float cores)
{
	Algebra schedule = solution;
	// distance == 1 -> change all neighbors by +- Wi processors
	// distance == 0 -> change only one neighbor by 1 processor

	map<int, float> vector_wi = schedule.find<Vector<int, float> >("wi")->getValues();
	int b = (int)schedule.find<Scalar<float> >("b")->getValue();
	int p = (int)schedule.find<Scalar<float> >("p")->getValue();
	const Vector<int, float> *Wi = schedule.find<Vector<int, float> >("Wi");
	map<int, float> vector_Wi = Wi->getValues();
	int n = vector_wi.size();

	//AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("Wi"));
	//AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("wi"));

	int num_tasks = ((int)((float)rand() / (float)RAND_MAX * n * tasks));
	//num_tasks = ((int)((float)RAND_MAX / (float)RAND_MAX * n * tasks));
	//num_tasks = 1; // Remove to restore randomness
	while (num_tasks > 0 && vector_Wi.size() > 0)
	{
		int size = vector_Wi.size();
		map<int, float>::iterator itask = vector_Wi.begin();
		int index = rand() % size;
		//index = 4; // Remove to restore randomness

		for(int i = 0; i < index; i++)
		{
			itask++;
		}

		// If a task is sequential, there is no neighbor to find there, remove that task from list and continue
		if(itask->second < 2)
		{
			vector_Wi.erase(itask);
			continue;
		}

		// Task maximal width > 1 (Wit > 0)
		int task = itask->first;
		int Wit = (int)vector_Wi[task];
		Wit = Wit > p ? p : Wit;
		Wit = (int)(log(Wit) / log(b));

		// task current allocated width
		int wi = (int)(log(vector_wi.find(task)->second) / log(b));

		//int amplitude = (Wit - 1) * cores;
		int amplitude = Wit * (int)cores; // + 1 to count the number of possibilities, -1 to remove the current allocation from possible new allocation
		amplitude = amplitude < 1 ? 1 : amplitude; // Make sure at least one transformation is performed

		//cerr << "# Task " << task << "; p = " << p << "; Wi = " << Wit << "; log2(Wi) = " << log(Wit) / log(b) << "; log2(wi) = " << wi << "; amplitude = " << amplitude << endl;

		// Count how many spots are available below and above current wi
		int avail_lower = wi;
		int avail_higher = Wit - wi;
	
		// Cound how many spots we need below and above current wi
		int wi_lower, wi_higher;
		if(rand() % 2 == 0)
		{
			wi_lower = (int)floor((float)amplitude / 2);
			wi_higher = (int)ceil((float)amplitude / 2);
		}
		else
		{
			wi_lower = (int)ceil((float)amplitude / 2);
			wi_higher = (int)floor((float)amplitude / 2);
		}
		
		int len_lower, len_higher;
		if(avail_lower >= wi_lower && avail_higher >=wi_higher)
		{
			len_lower = wi_lower;
			len_higher = wi_higher;
		}
		else if(avail_lower < wi_lower)
		{
			len_lower = avail_lower;
			len_higher = wi_higher + (wi_lower - avail_lower);
		}
		else if(avail_higher < wi_higher)
		{
			len_higher = avail_higher;
			len_lower = wi_lower + (wi_higher - avail_higher);
		}
		else
		{
			// Not enough space to fit the amplitude, aborting
			cerr << "# Not enough allocation opportunities to fit the amplitude" << endl;
			cerr << "# Task " << task << "; p = " << p << "; log2(Wi) = " << Wit << "; log2(wi) = " << wi << "; amplitude = " << amplitude << "; avail_lower = " << avail_lower << "; avail_higher = " << avail_higher << "; wi_lower = " << wi_lower << "; wi_higher = " << wi_higher << endl; 
			abort();
		}

		int alternative = ((int)rand() % (amplitude)) + 1; // Start counting the new allocation number with 1
		//alternative = 1; // Remove to restore randomness

		int new_wi = 0;
		if(alternative <= len_lower)
		{
			new_wi = wi - (len_lower - alternative + 1);
		}
		else if(alternative <= len_lower + len_higher)
		{
			new_wi = wi + alternative - len_lower;
		}
		else
		{
			cerr << "# Chosen alternative is beyond the spots available" << endl;
			cerr << "# Task " << task << "; p = " << p << "; log2(Wi) = " << Wit << "; log2(wi) = " << wi << "; amplitude = " << amplitude << "; avail_lower = " << avail_lower << "; avail_higher = " << avail_higher << "; wi_lower = " << wi_lower << "; wi_higher = " << wi_higher << "; alternative = " << alternative << endl; 
			len_lower = avail_lower;
			len_higher = wi_higher + (wi_lower - avail_lower);
		}
			
		if(wi == new_wi || new_wi > Wit || new_wi < 0)
		{
			cerr << "# Task " << task << "; Wi " << Wit << "; wi " << wi << "; amplitude " << amplitude << "; new_wi " << new_wi << endl;
			abort();
		}

		new_wi = (int)pow((double)b, (double)new_wi);
		if(new_wi == 64)
		{
			cerr << "# Task " << task << "; Wi " << Wit << "; wi " << wi << "; amplitude " << amplitude << "; new_wi " << new_wi << endl;
		}
		vector_wi[task] = new_wi;

		// Done with this task, continue with the next
		vector_Wi.erase(itask);
		num_tasks--;

	}

	Vector<int, float> wi("wi", vector_wi);
	//AmplOutput().dump(cerr, &wi);
	//cerr << "# =========================================" << endl << endl;
	//abort();
	schedule.insert(&wi);

	return schedule;
}

Algebra
neighbor_allocation_full(const Algebra &solution, float distance)
{
	return neighbor_allocation(solution, distance, distance);
}

Algebra
neighbor_allocation_tasks(const Algebra &solution, float distance)
{
	return neighbor_allocation(solution, distance, 0);
}

Algebra
neighbor_allocation_cores(const Algebra &solution, float distance)
{
	return neighbor_allocation(solution, 0, distance);
}

Algebra
neighbor_efficiency(const Algebra &solution, float distance)
{
	Algebra schedule = solution;
	
	const Scalar<float> *gemin = schedule.find<Scalar<float> >("gemin");
	if(gemin == NULL)
	{
		gemin = new Scalar<float>("gemin", 0);
		schedule.insert(gemin);
		delete gemin;
		gemin = schedule.find<Scalar<float> >("gemin");
	}
	float gemin_val = gemin->getValue();

	const Scalar<float> *gemax = schedule.find<Scalar<float> >("gemax");
	if(gemax == NULL)
	{
		gemax = new Scalar<float>("gemax", 1);
		schedule.insert(gemax);
		delete gemax;
		gemax = schedule.find<Scalar<float> >("gemax");
	}
	float gemax_val = gemax->getValue();

	// The whole universe we can vary the global minimal efficiency in 
	float universe = gemax_val - gemin_val;

	const Scalar<float> *geuse = schedule.find<Scalar<float> >("geuse");
	float geuse_val = 0;
	if(geuse == NULL)
	{
		// The initial value is somewhere between gemin and gemax ( = gemin + rand() * universe )
		geuse = new Scalar<float>("geuse", gemin_val + (universe * ((float)rand() / RAND_MAX)));
		schedule.insert(geuse);
		delete geuse;
		geuse = schedule.find<Scalar<float> >("geuse");
	}
	else
	{
		geuse_val = geuse->getValue();
		float random = (float)rand() / RAND_MAX;
		geuse_val = geuse_val + (random * universe * distance);

		if(geuse_val < gemin_val)
		{
			geuse_val += gemin_val - geuse_val;
		}
		else if(geuse_val > gemax_val)
		{
			geuse_val -= geuse_val - gemax_val;
		}

		if (geuse_val < gemin_val || geuse_val > gemax_val)
		{
			// Something went terribly wrong
			cerr << "Geuse val went off the chart despite (or because of) the out-of-universe problem fix" << endl;
			cerr << "geuse_val = " << geuse_val << "; gemin = " << gemin_val << "; gemax = " << gemax_val << "; universe = " << universe << "; random = " << random << ";distance = " << distance << endl; 
		}
	}

	// Reallocate
	return allocation_gemin(schedule, geuse_val);
	
	// This code focuses on task-individual efficiency. We target global efficiency	
#if 0
	Algebra schedule = solution;
	int granularity = 1000;

	// distance == 1 -> change all neighbors by +- Wi processors
	// distance == 0 -> change only one neighbor by 1 processor

	map<int, int> vector_wi = schedule.find<Vector<int, float> >("wi")->getValues();
	map<int, map<int, float> > matrix_e = schedule.find<Matrix<int, int, float> >("e")->getValues();
	int b = schedule.find<Scalar<float> >("b")->getValue();
	int p = schedule.find<Scalar<float> >("p")->getValue();
	const Vector<int, float> *Wi = schedule.find<Vector<int, float> >("Wi");
	map<int, int> vector_Wi = Wi->getValues();
	int n = vector_wi.size();

	// Extract the global threshold scalar, if any, or build one
	const Scalar<float> *gemin = schedule.find<Scalar<float> >("gemin");
	if(gemin == NULL)
	{
		gemin = new Scalar<float>("gemin", 0.5);
		schedule.insert(gemin);
		delete gemin;
		gemin = schedule.find<Scalar<float> >("gemin");
	}
	
	// Extract the {t}ask {e}efficiency {min}imum vector, if any, or build one
	const Vector<int, float> *temin = schedule.find<Vector<int, float> >("temin");
	if(temin == NULL)
	{
		map<int, float> t;
		for(map<int, int>::iterator iter = vector_wi.begin(); iter != vector_wi.end(); iter++)
		{
			//cerr << "Task " << iter->first << ", efficiency = " << global_threshold->getValue() << endl;
			t.insert(pair<int, float>(iter->first, gemin->getValue()));
		}

		temin = new Vector<int, float>("temin", t);
		schedule.insert(temin);
		delete temin;
		temin = schedule.find<Vector<int, float> >("temin");
	}
	map<int, float> map_temin = temin->getValues();

	// Extract the minimal efficiency this task accepts
	const Vector<int, float> *emin = schedule.find<Vector<int, float> >("emin");
	if(emin == NULL)
	{
		map<int, float> t;
		for(map<int, int>::iterator iter = vector_wi.begin(); iter != vector_wi.end(); iter++)
		{
			//cerr << "Task " << iter->first << ", efficiency = " << global_threshold->getValue() << endl;
			int task = iter->first;
			float min_e = 1;
			map<int, float> alle = matrix_e.find(task)->second;
			int maxp = 1;
			for(map<int, float>::const_iterator e = alle.begin(); e != alle.end() && maxp <= p; e++)
			{
				if(e->second < min_e)
				{
					min_e = e->second;
				}
				maxp++;
			}
			
			if(map_temin[task] < min_e)
			{
				map_temin[task] = min_e;
			}
			t.insert(pair<int, float>(task, min_e));
		}

		emin = new Vector<int, float>("emin", t);
		schedule.insert(emin);
		delete emin;
		emin = schedule.find<Vector<int, float> >("emin");
	}
	map<int, float> map_emin = emin->getValues();
	AmplOutput().dump(cerr, emin);
	
	// Extract the {r}elative {t}ask {e}efficiency {min}imum vector, if any, or build one
	const Vector<int, float> *rtemin = schedule.find<Vector<int, float> >("rtemin");
	if(rtemin == NULL)
	{
		map<int, float> t;
		for(map<int, int>::iterator iter = vector_wi.begin(); iter != vector_wi.end(); iter++)
		{
			//cerr << "Task " << iter->first << ", efficiency = " << global_threshold->getValue() << endl;
			int task = iter->first;
			t.insert(pair<int, float>(task, (1 - map_emin[task]) * gemin->getValue()));
		}

		rtemin = new Vector<int, float>("rtemin", t);
		schedule.insert(rtemin);
		delete rtemin;
		rtemin = schedule.find<Vector<int, float> >("rtemin");
	}
	map<int, float> map_rtemin = rtemin->getValues();

	//AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("Wi"));
	//AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("wi"));

	int num_tasks = ((int)((float)rand() / (float)RAND_MAX * n * distance));
	num_tasks = num_tasks == 0 ? 1 : num_tasks;
	//num_tasks = ((int)((float)RAND_MAX / (float)RAND_MAX * n * distance));
	//num_tasks = 1; // Remove to restore randomness
	while (num_tasks > 0 && vector_Wi.size() > 0)
	{
		int size = vector_Wi.size();
		map<int, int>::iterator itask = vector_Wi.begin();
		int index = rand() % size;
		//index = 4; // Remove to restore randomness

		for(int i = 0; i < index; i++)
		{
			itask++;
		}

		// If a task is sequential, there is no neighbor to find there, remove that task from list and continue
		if(itask->second < 2)
		{
			vector_Wi.erase(itask);
			continue;
		}

		// Task maximal width > 1 (Wit > 0)
		int task = itask->first;
		int max_efficiency = granularity;

		// task current allocated width
		int efficiency = map_rtemin.find(task)->second * granularity;
		cerr << "# Task " << task << ", efficiency = " << efficiency << endl;

		//int amplitude = (Wit - 1) * distance;
		int amplitude = max_efficiency * distance; // + 1 to count the number of possibilities, -1 to remove the current allocation from possible new allocation
		amplitude = amplitude < 1 ? 1 : amplitude; // Make sure at least one transformation is performed

		//cerr << "# Task " << task << "; p = " << p << "; Wi = " << Wit << "; log2(Wi) = " << log(Wit) / log(b) << "; log2(wi) = " << wi << "; amplitude = " << amplitude << endl;
		cerr << "# Task " << task << "; p = " << p << "; max efficiency = " << max_efficiency << "; efficiency = " << efficiency << "; amplitude = " << amplitude << endl; 

		// Count how many spots are available below and above current wi
		int avail_lower = efficiency;
		int avail_higher = max_efficiency - efficiency - 1;
	
		// Cound how many spots we need below and above current wi
		int eff_lower, eff_higher;
		if(rand() % 2 == 0)
		{
			eff_lower = floor((float)amplitude / 2);
			eff_higher = ceil((float)amplitude / 2);
		}
		else
		{
			eff_lower = ceil((float)amplitude / 2);
			eff_higher = floor((float)amplitude / 2);
		}
		
		int len_lower, len_higher;
		if(avail_lower >= eff_lower && avail_higher >= eff_higher)
		{
			len_lower = eff_lower;
			len_higher = eff_higher;
		}
		else if(avail_lower < eff_lower)
		{
			len_lower = avail_lower;
			len_higher = eff_higher + (eff_lower - avail_lower);
		}
		else if(avail_higher < eff_higher)
		{
			len_higher = avail_higher;
			len_lower = eff_lower + (eff_higher - avail_higher);
		}
		else
		{
			// Not enough space to fit the amplitude, aborting
			cerr << "# Not enough allocation opportunities to fit the amplitude" << endl;
			cerr << "# Task " << task << "; p = " << p << "; max efficiency = " << max_efficiency << "; efficiency = " << efficiency << "; amplitude = " << amplitude << "; avail_lower = " << avail_lower << "; avail_higher = " << avail_higher << "; eff_lower = " << eff_lower << "; eff_higher = " << eff_higher << endl; 
			abort();
		}

		int alternative = ((int)rand() % (amplitude)) + 1; // Start counting the new allocation number with 1
		//alternative = 1; // Remove to restore randomness

		int new_eff;
		if(alternative <= len_lower)
		{
			new_eff = efficiency - (len_lower - alternative + 1);
		}
		else if(alternative <= len_lower + len_higher)
		{
			new_eff = efficiency + alternative - len_lower;
		}
		else
		{
			cerr << "# Chosen alternative is beyond the spots available" << endl;
			cerr << "# Task " << task << "; p = " << p << "; max efficiency = " << max_efficiency << "; efficiency = " << efficiency << "; amplitude = " << amplitude << "; avail_lower = " << avail_lower << "; avail_higher = " << avail_higher << "; eff_lower = " << eff_lower << "; eff_higher = " << eff_higher << endl; 
			len_lower = avail_lower;
			len_higher = eff_higher + (eff_lower - avail_lower);
		}

		cerr << "# Task " << task << "; p = " << p << "; max efficiency = " << max_efficiency << "; efficiency = " << efficiency << "; amplitude = " << amplitude << "; avail_lower = " << avail_lower << "; avail_higher = " << avail_higher << "; eff_lower = " << eff_lower << "; eff_higher = " << eff_higher << "; new_eff = " << new_eff << endl; 
			
		if(efficiency == new_eff || new_eff > max_efficiency || new_eff < 0)
		{
			cerr << "# Task " << task << "; max efficiency " << max_efficiency << "; efficiency " << efficiency << "; amplitude " << amplitude << "; new_eff " << new_eff << endl;
			abort();
		}

		map_rtemin[task] = (float)new_eff / granularity;
		map_temin[task] = (float)map_emin[task] + (1 - map_emin[task]) * map_rtemin[task];

		// Done with this task, continue with the next
		vector_Wi.erase(itask);
		num_tasks--;

	}

	Vector<int, float> new_rtemin("rtemin", map_rtemin);
	Vector<int, float> new_temin("temin", map_temin);
	AmplOutput().dump(cerr, &new_rtemin);
	AmplOutput().dump(cerr, &new_temin);
	AmplOutput().dump(cerr, gemin);
	schedule.insert(&new_rtemin);
	schedule.insert(&new_temin);

	AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("wi"));
	schedule = allocation_temin(schedule);
	AmplOutput().dump(cerr, schedule.find<Vector<int, float> >("wi"));
	cerr << "# =========================================" << endl << endl;

	return schedule;
#endif
}

float
quality_simple(const Algebra &schedule)
{
	//int n = schedule.find<Scalar<float> >("n")->getValue();
	int p = (int)schedule.find<Scalar<float> >("p")->getValue();
	float alpha = schedule.find<Scalar<float> >("alpha")->getValue();
	float zeta = schedule.find<Scalar<float> >("zeta")->getValue();
	float M = schedule.find<Scalar<float> >("M")->getValue();
	Vector<int, float> tau = schedule.find<Vector<int, float> >("Tau");
	Vector<int, float> Wi = schedule.find<Vector<int, float> >("Wi");
	Vector<int, float> wi = schedule.find<Vector<int, float> >("wi");
	Vector<int, float> frequency = schedule.find<Vector<int, float> >("frequency");
	Vector<int, float> group = schedule.find<Vector<int, float> >("group");
	Matrix<int, int, float> e = schedule.find<Matrix<int, int, float> >("e");
	
	double static_energy = M * p;
	double dynamic_energy = 0;

	// Now evaluate the quality of this schedule
	for(map<int, float>::const_iterator i = group.getValues().begin(); i != group.getValues().end(); i++)
	{
		int task = i->first;

		float tau_t = tau.find(task);
		int width_t = (int)wi.find(task);
		//double e_wt = e.find(width_t, task);
		width_t = width_t > p ? p : width_t;
		int f_t = (int)frequency.find(task);

		double area = (double)tau_t / (double)(e.find(width_t, task)) / (double)frequency.find(task);
		static_energy -= area;
		dynamic_energy += area * pow(f_t, alpha);
	}

	return dynamic_energy + zeta * static_energy;
}

static
float
energy_static_idle(const Algebra &schedule, float epsilon)
{
	//int n = schedule.find<Scalar<float> >("n")->getValue();
	int p = (int)schedule.find<Scalar<float> >("p")->getValue();
	float alpha = schedule.find<Scalar<float> >("alpha")->getValue();
	float zeta = schedule.find<Scalar<float> >("zeta")->getValue();
	float minF = *(schedule.find<Set<float> >("F")->getValues().begin());
	
	float M;
	const Scalar<float> *float_M = schedule.find<Scalar<float> >("M");
	if(float_M != NULL)
	{
		M = float_M->getValue();
	}
	else
	{
		const Scalar<float> *int_M = schedule.find<Scalar<float> >("M");
		M = int_M->getValue();
	}

	Vector<int, float> tau = schedule.find<Vector<int, float> >("Tau");
	Vector<int, float> Wi = schedule.find<Vector<int, float> >("Wi");
	Vector<int, float> wi = schedule.find<Vector<int, float> >("wi");
	Vector<int, float> frequency = schedule.find<Vector<int, float> >("frequency");
	Vector<int, float> group = schedule.find<Vector<int, float> >("group");
	Matrix<int, int, float> e = schedule.find<Matrix<int, int, float> >("e");
	
	double idle_energy = 0;
	double busy_energy = 0;
	double total_area = 0;

	// Now evaluate the quality of this schedule
	for(map<int, float>::const_iterator i = group.getValues().begin(); i != group.getValues().end(); i++)
	{
		int task = i->first;

		float tau_t = tau.find(task);
		float width_t = wi.find(task);
		width_t = width_t > p ? p : width_t;
		//double e_wt = e.find(width_t, task);
		float f_t = frequency.find(task);

		double area = (double)tau_t / (double)(e.find((int)width_t, task)) / (double)frequency.find(task);

		total_area += area;
		busy_energy += area * pow(f_t, alpha) + zeta * f_t;
	}
	idle_energy = (p * M - total_area) * pow(minF, alpha) + zeta * minF;

	return busy_energy + epsilon * idle_energy;
}

float
quality_generic(const Algebra &schedule)
{
	//int n = schedule.find<Scalar<float> >("n")->getValue();
	int p = (int)schedule.find<Scalar<float> >("p")->getValue();
	float alpha = schedule.find<Scalar<float> >("alpha")->getValue();
	float alpha_static = schedule.find<Scalar<float> >("alpha-static")->getValue();
	float zeta = schedule.find<Scalar<float> >("zeta")->getValue();
	float epsilon = schedule.find<Scalar<float> >("epsilon")->getValue();
	float minF = *(schedule.find<Set<float> >("F")->getValues().begin());
	
	float M;
	const Scalar<float> *float_M = schedule.find<Scalar<float> >("M");
	if(float_M != NULL)
	{
		M = float_M->getValue();
	}
	else
	{
		const Scalar<float> *int_M = schedule.find<Scalar<float> >("M");
		M = int_M->getValue();
	}

	Vector<int, float> tau = schedule.find<Vector<int, float> >("Tau");
	Vector<int, float> Wi = schedule.find<Vector<int, float> >("Wi");
	Vector<int, float> wi = schedule.find<Vector<int, float> >("wi");
	Vector<int, float> frequency = schedule.find<Vector<int, float> >("frequency");
	Vector<int, float> group = schedule.find<Vector<int, float> >("group");
	Matrix<int, int, float> e = schedule.find<Matrix<int, int, float> >("e");
	
	double idle_energy = 0;
	double busy_energy = 0;
	double total_area = 0;

	// Now evaluate the quality of this schedule
	for(map<int, float>::const_iterator i = group.getValues().begin(); i != group.getValues().end(); i++)
	{
		int task = i->first;

		float tau_t = tau.find(task);
		float width_t = wi.find(task);
		width_t = width_t > p ? p : width_t;
		//double e_wt = e.find(width_t, task);
		float f_t = frequency.find(task);

		double area = (double)tau_t / (double)(e.find((int)width_t, task)) / (double)frequency.find(task);

		total_area += area;
		busy_energy += area * pow(f_t, alpha) + zeta * f_t;
	}
	idle_energy = (p * M - total_area) * pow(minF, alpha_static) + zeta * minF;

	return busy_energy + epsilon * idle_energy;
}

float
quality_idle(const Algebra &schedule)
{
	return energy_static_idle(schedule, 0.5);
}

float
quality_off(const Algebra &schedule)
{
	return energy_static_idle(schedule, 0);
}

#if RETURN_EARLY
#define return_now_if_best(solution) if(solution.find<Scalar<float> >("all_sequential")->getValue() == 1 && solution.find<Scalar<float> >("frequency_optimal")->getValue() == 1) return solution
#else
#define return_now_if_best(solution)
#endif

Algebra
annealing(const Algebra &initial, const Algebra &initial_best, float (*quality)(const Algebra &solution), Algebra (*neighbor)(const Algebra &solution, float distance), float temperature, float cooling, float final, int max_transform, int max_new_state, float distance)
{
	Algebra schedule = initial;

	return_now_if_best(initial_best);

	
	schedule = neighbor(schedule, distance);

	float m;
	int counter = 0;
	float T = temperature;

	Algebra best = initial_best;
	float best_q = quality(best);
	Vector<int, float> *best_solution = initial_best.find<Vector<int, float> >("wi")->clone();
	float q = quality(schedule);

	while (T > final)
	{
		int new_state = 0;

		for(int i = 0; i < max_transform; i++)
		{
			// Keep track of the solution of last iteration
			Vector<int, float> *old_solution = schedule.find<Vector<int, float> >("wi")->clone();
			float old_q = q;

			// Transform the solution
			schedule = neighbor(schedule, distance);
			schedule = mapping_ltlg(schedule);
			schedule = frequency_height(schedule);

			// Return immediately if this schedule cannot be possibly improved (all sequential tasks and all tasks running at lowest frequency)
			return_now_if_best(schedule);
			
			m = schedule.find<Scalar<float> >("m")->getValue();
			q = quality(schedule);
			
			// If the new energy is lower than the previous one, then we accept this solution
			if (q - old_q <= 0 && m >= 0)
			{
				delete old_solution;

				// If this solution is better than the best found so far, then keep this best solution
				if (q < best_q)
				{
					best_solution = schedule.find<Vector<int, float> >("wi")->clone();
					best_q = q;
				}

				// One new state validated
				new_state++;
			}
			else
			{
				// Randomly accept a degrading solution, more likely when the temperature is higher
				if (rand() / RAND_MAX <= exp(-(q - old_q) / T) && m >= 0)
				{
					// Update current solution
					delete old_solution;

					// One new state validated
					new_state++;
				}
				else
				{
					// Revert old solution 
					schedule.insert(old_solution);
					q = old_q;
				}
			}
			
			// Too much successful transformations
			if(new_state == max_new_state) 
			{
				break;
			}
		}
	
		T = T * cooling;
		counter++;
	}

	schedule.insert(best_solution);
	schedule = mapping_ltlg(schedule);
	schedule = frequency_height(schedule);

	return schedule;
}

float
annealing_complexity(float temperature, float cooling, float final, int max_transform, int max_new_state, float distance)
{
	return max_transform * ceil(log(temperature / final) / log(1 / cooling));
}

}
}
